package sim.listener;

import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import sim.math.SVector3d;
import sim.opengl.SMovableCamera;
import sim.util.SMouseEventHandler;

/**
 * La classe <b>SListenerCamera</b> est une camra qui implmente l'interface <b>MouseMotionListener</b> et <b>MouseListener</b>.
 * 
 * @author Simon Vezina
 * @since 2020-06-15
 * @version 2021-05-25
 */
public class SListenerCamera implements  MouseMotionListener, MouseListener {

  private static final SVector3d DEFALUT_CAMERA_POSITON = SVector3d.ORIGIN;
  private static final SVector3d DEFAULT_CAMERA_LOOK_AT = SVector3d.UNITARY_Z;
  private static final SVector3d DEFAULT_CAMERA_UP = SVector3d.UNITARY_X;
  
  private final SMovableCamera camera;
  private final SMouseEventHandler mouse;
  
  /**
   * Constructeur d'une camra par dfaut.
   */
  public SListenerCamera()
  {
    this(DEFALUT_CAMERA_POSITON, DEFAULT_CAMERA_LOOK_AT, DEFAULT_CAMERA_UP);
  }
  
  /**
   * Constructeur d'une camra.
   * 
   * @param position La position de la camera.
   * @param look_at La position o regarde la camera.
   * @param up L'orientation vers le haut de la camera.
   */
  public SListenerCamera(SVector3d position, SVector3d look_at, SVector3d up)
  {
    this.camera = new SMovableCamera(position, look_at, up);
    this.mouse = new SMouseEventHandler();
  }
  
  /**
   * ...
   * 
   * @return
   */
  public SMovableCamera getCamera()
  {
    return this.camera;
  }
  
  @Override
  public void mouseClicked(MouseEvent e) 
  {
    mouse.clicked(e);
  }

  @Override
  public void mousePressed(MouseEvent e) 
  {
    mouse.pressed(e);
  }

  @Override
  public void mouseReleased(MouseEvent e)
  {
    mouse.released(e);
  }

  @Override
  public void mouseEntered(MouseEvent e)
  {
        
  }

  @Override
  public void mouseExited(MouseEvent e)
  {
        
  }

  @Override
  public void mouseDragged(MouseEvent e)
  {
    mouse.dragged(e);
    
    if(mouse.isButtonPressed(1) && mouse.isButtonPressed(3))
      camera.rotationUp(mouse.getDeltaY(), mouse.getDeltaX());
    else
    {
      if(mouse.isButtonPressed(1))
        camera.rotationLookAtOrientation(mouse.getDeltaX(), -1.0*mouse.getDeltaY());
    
      if(mouse.isButtonPressed(3))
        camera.moveCamera(-1.0*mouse.getDeltaY(), mouse.getDeltaX());
    }
  }

  @Override
  public void mouseMoved(MouseEvent e) 
  {
    mouse.moved(e);
  }

}
